#!/usr/bin/env python
# -*- coding: utf-8 -*-
# =============================================================================
# Licensed under a 3-clause BSD style license, see LICENSE.
# =============================================================================
"""
Set of utilities for comparison of floating point numbers inspired by
- https://randomascii.wordpress.com/2012/02/25/comparing-floating-point-numbers-2012-edition/
and the previous version:
- http://www.cygnus-software.com/papers/comparingfloats/Obsolete%20comparing%20floating%20point%20numbers.htm
"""

__all__ = (
    "distance",  ## distance in ULPs
    "isclose",  ## close in  terms of ULPs ?
    "isequal",  ## isequal using  ULPs and SCALE
    "next_double",  ## get next double
)


def distance(a, b):
    """
    Distance in ULPS between two (floating point) numbers.
    It is assumed here that  size(long)==size(double) for underlying C-library!

    Example
    -------
    >>> a = ...
    >>> b = ...
    >>> print distance ( a , b )
    """
    if a == b:
        return 0
    elif a > b:
        return -distance(b, a)
    elif b <= 0:
        return distance(-b, -a)
    elif a < 0:
        return distance(0, -a) + distance(0, b)

    ## here a and b have same sign

    import ctypes

    a, b = abs(a), abs(b)

    aa = ctypes.c_double(float(a))
    bb = ctypes.c_double(float(b))

    al = ctypes.c_long.from_buffer(aa).value
    bl = ctypes.c_long.from_buffer(bb).value

    return bl - al


def next_double(a, ulps=1):
    """
    Get the ``next-double'' by certain ULPs distance.
    It is assumed here that size(long)==size(double) for underlying C-library!

    Examples
    --------
    >>> a = next_double ( 1 , 1 )
    >>> print a , a - sys.float_info.epsilon
    """

    a = float(a)
    if 0 == ulps:
        return a
    elif a < 0:
        return -next_double(-a, -ulps)
    elif 0 > ulps:
        d = distance(a, 0.0) + ulps
        if d < 0:
            return -next_double(0.0, -d)

    ## aply a trick  with casting

    import ctypes

    aa = ctypes.c_double(a)
    al = ctypes.c_long.from_buffer(aa).value
    al = ctypes.c_long(al + ulps)
    aa = ctypes.c_double.from_buffer(al).value

    return aa


def isclose(a, b, ulps=1000):
    """
    Are two floating point numbers close enough (in units is ULPs)?

    Examples
    --------
    >>> a = ...
    >>> b = ...
    >>> print isclose ( a , b , 1000 )
    """
    return (a == b) or ulps >= abs(distance(a, b))


def isequal(a, b, scale=1.0, absdiff=0.0, ulps=1000):
    """
    Are  two numbers ``a'' and ``b''  close enough ?
    Numbers are considered to be equal is :
    - (for absdiff >0) the absolute difference is smaller than absdiff
    - *OR* they are close enough in ULPs
    - *OR* (for scale != 0) the difference is small enough compared to the scale

    Examples
    --------
    >>> isequal(1,1+sys.float_info.epsilon)
    True
    >>> isequal(1,1+sys.float_info.epsilon, ulps = 0 )
    False
    >>> isequal(1,1+sys.float_info.epsilon, scale = 100 , ulps = 0 )
    True
    """
    return (
        (a == b)
        or (0 < absdiff and abs(a - b) < absdiff)
        or isclose(a, b, ulps)
        or (scale and isclose((a - b) + scale, scale, ulps))
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license, see LICENSE.
"""
Tests for the skhep.math.kinematics module.
"""

# -----------------------------------------------------------------------------
# Import statements
# -----------------------------------------------------------------------------
import pytest
from pytest import approx

from hepunits.units import GeV, ps
from hepunits.constants import hbar

from skhep.math.kinematics import *
from skhep.math.vectors import Vector3D

# -----------------------------------------------------------------------------
# Actual tests
# -----------------------------------------------------------------------------


def test_kinematics_Kallen_function():
    assert Kallen_function(2, 1, 0) == 1


def test_valid_Armenteros_Podolanski_variables():
    d1 = Vector3D(1.0, 2.0, 3.0)
    d2 = Vector3D(1.0, -2.0, 3.0)
    assert Armenteros_Podolanski_variables(d1, d2) == (2.0, 0.0)


def test_invalid_Armenteros_Podolanski_variables():
    d1 = Vector3D(1.0, 2.0, 3.0)
    d2 = Vector3D(-1.0, -2.0, -3.0)
    with pytest.raises(ValueError):
        Armenteros_Podolanski_variables(d1, d2)


def test_width_lifetime_conversions():
    assert lifetime_to_width(1.5 * ps) / GeV == approx(4.388079676311604e-13)
    assert 1.5 * ps * lifetime_to_width(1.5 * ps) == hbar
    assert width_to_lifetime(hbar) == 1 * MeV
    #
    with pytest.raises(ValueError):
        lifetime_to_width(-1)
    with pytest.raises(ValueError):
        width_to_lifetime(-1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license, see LICENSE.
"""
Tests for the skhep.math.geometry module.
"""

# -----------------------------------------------------------------------------
# Import statements
# -----------------------------------------------------------------------------
from skhep.math.geometry import *

import pytest
from pytest import approx

from math import pi


# -----------------------------------------------------------------------------
# Actual tests
# -----------------------------------------------------------------------------
def test_geometry_constructors():
    v1 = Vector3D()
    p1 = Point3D()
    assert str(p1) == "Point3D(x=0.0,y=0.0,z=0.0)"
    assert repr(p1) == "<Point3D(x=0.0,y=0.0,z=0.0)>"

    l1 = Line3D(p1, Vector3D(0, 0, 1))
    assert str(l1) == "Line3D({0},{1})".format(p1, Vector3D(0, 0, 1))
    assert repr(l1) == "<Line3D({0},{1})>".format(p1, Vector3D(0, 0, 1))

    pl1 = Plane3D(p1, Vector3D(0, 0, 1))
    assert str(pl1) == "Plane3D({0},{1})".format(p1, Vector3D(0, 0, 1))
    assert repr(pl1) == "<Plane3D({0},{1})>".format(p1, Vector3D(0, 0, 1))

    l2 = Line3D.from_points(p1, Point3D(1, 2, 3))
    pl2 = Plane3D.from_points(p1, Point3D(1, 2, 3), Point3D(3, 2, 1))
    p2 = Point3D.frompoint(1, 1, 1)
    pl3 = Plane3D.from_line_and_point(l1, p2)

    p3 = Point3D.fromiterable([1.0, 1.0, 1.0])
    assert p3 == Point3D(1.0, 1.0, 1.0)
    p4 = Point3D.fromcylindricalcoords(1.0, 0.0, 1.0)
    assert p4 == Point3D(1.0, 0.0, 1.0)
    assert p4.rho == 1.0
    p5 = Point3D.fromsphericalcoords(1.0, 0.0, 0.0)
    assert p5 == Point3D(0.0, 0.0, 1.0)
    assert p5.theta == 0.0
    assert p5.phi == 0.0

    v2 = Vector3D(1, 1, 1)
    with pytest.raises(NotImplementedError):
        Line3D.__init__(Line3D(), p1, p1)
    with pytest.raises(NotImplementedError):
        Line3D.__init__(Line3D(), v2, v2)
    with pytest.raises(ValueError):
        Line3D.__init__(Line3D(), p1, v1)

    with pytest.raises(NotImplementedError):
        Plane3D.__init__(Plane3D(), p1, p1)
    with pytest.raises(NotImplementedError):
        Plane3D.__init__(Plane3D(), v2, v2)
    with pytest.raises(ValueError):
        Plane3D.__init__(Plane3D(), p1, v1)


def test_operators():
    v1 = Vector3D(1, 2, 3)
    v2 = Vector3D(3, 2, 1)
    p1 = Point3D(0, 0, 1)
    p2 = Point3D(0, 1, 0)
    line1 = Line3D(p1, v1)
    line2 = Line3D(p2, v2)
    plane1 = Plane3D(p1, v1)
    plane2 = Plane3D(p2, v2)

    assert p1 + v1 == Point3D(1, 2, 4)
    assert v1 + p1 == Point3D(1, 2, 4)

    assert p1 - v1 == Point3D(-1, -2, -2)
    assert p1 - p2 == Vector3D(0, -1, 1)

    p1 -= v1
    p1 += v1

    assert p1 == p1
    assert p1 != p2

    assert line1 == line1
    assert line1 != line2

    assert plane1 == plane1
    assert plane1 != line2


def test_contains():
    line = Line3D(Point3D(), Vector3D(0, 0, 1))
    plane = Plane3D(Point3D(), Vector3D(0, 0, 1))

    assert Point3D(0, 0, -10) in line
    assert Point3D(1, 0, -10) not in line

    assert Point3D(1, 0, 0) in plane
    assert Point3D(0, 1, 0) in plane
    assert Point3D(0, 0, 2) not in plane

    linex = Line3D(Point3D(), Vector3D(1, 0, 0))
    liney = Line3D(Point3D(), Vector3D(0, 1, 0))
    linez = Line3D(Point3D(), Vector3D(0, 0, 1))

    assert linex in plane
    assert liney in plane
    assert linez not in plane


def test_distance():
    with pytest.raises(NotImplementedError):
        Point3D.distance(Point3D(), 2.0)
    with pytest.raises(NotImplementedError):
        Line3D.distance(Line3D(), 2.0)
    with pytest.raises(NotImplementedError):
        Plane3D.distance(Plane3D(), 2.0)

    p0 = Point3D()
    p1 = Point3D(1, 0, 0)

    assert p0.distance(p0) == 0
    assert p0.distance(p1) == 1

    line = Line3D(Point3D(), Vector3D(0, 0, 1))

    assert p0.distance(line) == 0
    assert p1.distance(line) == 1

    assert line.distance(p0) == 0
    assert line.distance(p1) == 1

    plane = Plane3D(Point3D(1, 1, 0), Vector3D(0, 0, 1))

    assert plane.distance(plane) == 0
    assert plane.distance(p0) == 0

    line1 = Line3D(Point3D(1, 2, 3), Vector3D(4, 5, 6))
    line2 = Line3D(Point3D(0, 0, 1), Vector3D(1, 1, 0))

    assert plane.distance(line1) == 0
    assert plane.distance(line2) == 1
    assert line1.distance(plane) == 0
    assert line2.distance(plane) == 1

    line3 = Line3D(Point3D(0, 0, 2), Vector3D(1, 1, 0))

    assert line3.distance(line2) == 1


def test_angle():
    p0 = Point3D()

    with pytest.raises(NotImplementedError):
        Line3D.angle(Line3D(), p0)
    with pytest.raises(NotImplementedError):
        Plane3D.angle(Plane3D(), p0)

    line0 = Line3D(Point3D(), Vector3D(0, 0, 1))
    line1 = Line3D(Point3D(0, 0, 1), Vector3D(0, 1, 0))

    plane0 = Plane3D(Point3D(), Vector3D(0, 0, 1))
    plane1 = Plane3D(Point3D(0, 0, 1), Vector3D(0, 1, 0))

    assert line0.angle(line1) == pi / 2.0
    assert line1.angle(line0) == pi / 2.0
    assert line1.angle(line1) == 0.0
    assert line0.angle(plane0) == 0.0
    assert line0.angle(plane1) == pi / 2.0
    assert line1.angle(plane0) == pi / 2.0
    assert line1.angle(plane1) == 0.0

    assert plane0.angle(plane1) == pi / 2.0
    assert plane1.angle(plane0) == pi / 2.0
    assert plane1.angle(plane1) == 0.0

    assert plane0.angle(line0) == 0.0
    assert plane0.angle(line1) == pi / 2.0
    assert plane1.angle(line0) == pi / 2.0
    assert plane1.angle(line1) == 0.0


def test_intersect():
    with pytest.raises(NotImplementedError):
        Line3D.intersect(Line3D(), 1)
    with pytest.raises(NotImplementedError):
        Plane3D.intersect(Plane3D(), 1)

    p0 = Point3D()
    p1 = Point3D(0, 0, 1)

    line0 = Line3D(Point3D(), Vector3D(0, 0, 1))
    line1 = Line3D(Point3D(0, 0, 1), Vector3D(0, 1, 0))

    plane0 = Plane3D(Point3D(), Vector3D(0, 0, 1))
    plane1 = Plane3D(Point3D(0, 0, 1), Vector3D(0, 1, 0))

    assert line0.intersect(p0) == p0
    assert line0.intersect(p1) == p1
    assert line1.intersect(p0) == None
    assert line1.intersect(p1) == p1

    assert line0.intersect(line1) == Point3D(0, 0, 1)
    assert line1.intersect(line0) == Point3D(0, 0, 1)

    assert line0.intersect(plane0) == Point3D()
    assert line0.intersect(plane1) == line0
    assert line1.intersect(plane0) == None
    assert line1.intersect(plane1) == Point3D(0, 0, 1)

    assert plane0.intersect(p0) == p0
    assert plane0.intersect(p1) == None
    assert plane1.intersect(p0) == p0
    assert plane1.intersect(p1) == p1

    assert plane0.intersect(line0) == Point3D()
    assert plane0.intersect(line1) == None
    assert plane1.intersect(line0) == line0
    assert plane1.intersect(line1) == Point3D(0, 0, 1)

    assert plane0.intersect(plane1) == Line3D(Point3D(), Vector3D(1, 0, 0))
    assert plane1.intersect(plane0) == Line3D(Point3D(), Vector3D(1, 0, 0))

    assert plane0.intersect(plane0) == None
    line2 = Line3D(Point3D(0, 1, 0), Vector3D(0, 0, 1))
    assert line2.intersect(line0) == None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# =============================================================================
# copied from https://raw.githubusercontent.com/PythonCHB/close_pep/master/is_close.py
# =============================================================================
"""
Test implementation for an isclose() function, for possible inclusion in
the Python standard library -- PEP0485

This version has multiple methods in it for experimentation and testing.

The ``final'' version can be found in isclose.py

This implementation is the result of much discussion on the python-ideas list
in January, 2015:

   https://mail.python.org/pipermail/python-ideas/2015-January/030947.html

   https://mail.python.org/pipermail/python-ideas/2015-January/031124.html

   https://mail.python.org/pipermail/python-ideas/2015-January/031313.html

Copyright: Christopher H. Barker
License: Apache License 2.0 http://opensource.org/licenses/apache2.0.php

"""
__all__ = ("isclose",)  ## close using relative and absolute distances

import cmath


def isclose(a, b, rel_tol=1e-9, abs_tol=0.0, method="weak"):
    """
    returns True if a is close in value to b. False otherwise

    :param a: one of the values to be tested

    :param b: the other value to be tested

    :param rel_tol=1e-8: The relative tolerance -- the amount of error
                         allowed, relative to the magnitude of the input
                         values.

    :param abs_tol=0.0: The minimum absolute tolerance level -- useful for
                        comparisons to zero.

    :param method: The method to use. options are:
                  "asymmetric" : the b value is used for scaling the tolerance
                  "strong" : The tolerance is scaled by the smaller of
                             the two values
                  "weak" : The tolerance is scaled by the larger of
                           the two values
                  "average" : The tolerance is scaled by the average of
                              the two values.

    NOTES:

    -inf, inf and NaN behave similar to the IEEE 754 standard. That
    -is, NaN is not close to anything, even itself. inf and -inf are
    -only close to themselves.

    Complex values are compared based on their absolute value.

    The function can be used with Decimal types, if the tolerance(s) are
    specified as Decimals::

      isclose(a, b, rel_tol=Decimal('1e-9'))

    See PEP-0485 for a detailed description

    """
    if method not in ("asymmetric", "strong", "weak", "average"):
        raise ValueError(
            'method must be one of: "asymmetric",' ' "strong", "weak", "average"'
        )

    if rel_tol < 0.0 or abs_tol < 0.0:
        raise ValueError("error tolerances must be non-negative")

    if a == b:  # short-circuit exact equality
        return True
    # use cmath so it will work with complex or float
    if cmath.isinf(a) or cmath.isinf(b):
        # This includes the case of two infinities of opposite sign, or
        # one infinity and one finite number. Two infinities of opposite sign
        # would otherwise have an infinite relative tolerance.
        return False
    diff = abs(b - a)
    if method == "asymmetric":
        return (diff <= abs(rel_tol * b)) or (diff <= abs_tol)
    elif method == "strong":
        return ((diff <= abs(rel_tol * b)) and (diff <= abs(rel_tol * a))) or (
            diff <= abs_tol
        )
    elif method == "weak":
        return ((diff <= abs(rel_tol * b)) or (diff <= abs(rel_tol * a))) or (
            diff <= abs_tol
        )
    elif method == "average":
        return diff <= abs(rel_tol * (a + b) / 2) or (diff <= abs_tol)
    else:
        raise ValueError(
            "method must be one of:" ' "asymmetric", "strong", "weak", "average"'
        )


# =============================================================================
## The END
# =============================================================================



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# =============================================================================
# Licensed under a 3-clause BSD style license, see LICENSE.
# =============================================================================
"""
Geometry classes
================

Three geometry classes are available:
* ``Point3D``      : a point in 3-dimensional space
* ``Line3D``       : a line  in 3-dimensional space
* ``Plane3D``      : a plane in 3-dimensional space

"""
# =============================================================================
__all__ = (
    "Vector3D",  ## 3D-vector
    "Point3D",  ## 3D-point
    "Line3D",  ## line in 3D
    "Plane3D",  ## plane in 3D
    "distance",  ## 3D-distance between geometry objects
)
# =============================================================================
# Import statements
# =============================================================================
from skhep.math.vectors import Vector3D
from skhep.math.numeric import isequal


# =============================================================================
# 3D-point
# =============================================================================
class Point3D(object):
    """Point in 3D space."""

    def __init__(self, *args, **kwargs):
        self._vct = Vector3D(*args, **kwargs)

    @classmethod
    def origin(cls):
        """Shortcut constuctor for the origin (x=0.,y=0.,z=0.).
        Equivalent to the default constructor Point3D().
        """
        return cls(0.0, 0.0, 0.0)

    @classmethod
    def frompoint(cls, x, y, z):
        """Constructor from an explicit space point."""
        return cls(x, y, z)

    @classmethod
    def fromvector(cls, other):
        """Copy constructor."""
        return cls(other.x, other.y, other.z)

    @classmethod
    def fromsphericalcoords(cls, r, theta, phi):
        """Constructor from a space point specified in spherical coordinates.

        Parameters
        ----------
        r     : radius, the radial distance from the origin (r > 0)
        theta : inclination in radians (theta in [0, pi] rad)
        phi   : azimuthal angle in radians (phi in [0, 2pi) rad)
        """
        from math import sin, cos

        x = r * sin(theta) * cos(phi)
        y = r * sin(theta) * sin(phi)
        z = r * cos(theta)
        return cls(x, y, z)

    @classmethod
    def fromcylindricalcoords(cls, rho, phi, z):
        """Constructor from a space point specified in cylindrical coordinates.

        Parameters
        ----------
        rho : radial distance from the z-axis (rho > 0)
        phi : azimuthal angle in radians (phi in [-pi, pi) rad)
        z   : height
        """
        from math import sin, cos

        x = rho * cos(phi)
        y = rho * sin(phi)
        z = z
        return cls(x, y, z)

    @classmethod
    def fromiterable(cls, values):
        """Constructor from a suitable iterable object.
        Suitable means here that all entries are numbers
        and the length equals 3.
        """
        values = list(values)
        if not len(values) == 3:
            raise ValueError(
                "Input iterable length = {0}! Please check your inputs.".format(
                    len(values)
                )
            )
        for i, v in enumerate(values):
            if not isinstance(v, (int, float)):
                raise ValueError("Component #{0} is not a number!".format(i))
        return cls(values[0], values[1], values[2])

    @property
    def x(self):
        """Return the x, aka first coordinate at position 0."""
        return self._vct.x

    @x.setter
    def x(self, value):
        """Sets x, aka first coordinate at position 0."""
        self._vct.x = value

    @property
    def y(self):
        """Return the y, aka second coordinate at position 1."""
        return self._vct.y

    @y.setter
    def y(self, value):
        """Sets y, aka second coordinate at position 1."""
        self._vct.y = value

    @property
    def z(self):
        """Return the z, aka third coordinate at position 2."""
        return self._vct.z

    @z.setter
    def z(self, value):
        """Sets z, aka third coordinate at position 2."""
        self._vct.z = value

    @property
    def rho(self):
        """Return the cylindrical coordinate rho."""
        return self._vct.rho

    @property
    def theta(self, deg=False):
        """Return the spherical coordinate theta.

        Parameters
        ----------
        deg : float, optional
            Return the angle in degrees (default is radians).
        """
        return self._vct.theta(deg)

    @property
    def phi(self, deg=False):
        """Return the spherical or cylindrical coordinate phi.

        Parameters
        ----------
        deg : float, optional
            Return the angle in degrees (default is radians).
        """
        return self._vct.phi(deg)

    def copy(self):
        """Make a copy of this point."""
        return Point3D(self[0], self[1], self[2])

    def __setitem__(self, i, value):
        """Update/set the ith points component (commencing at 0, of course)."""
        self._vct[i] = value

    def __getitem__(self, i):
        """Get the ith point component (commencing at 0, of course)."""
        return self._vct[i]

    def tolist(self):
        """Return the vector as a list."""
        return self._vct.tolist()

    def __len__(self):
        """Length/dimension of the poit, i.e. the number of elements = 3."""
        return 3

    @property
    def mag(self):
        """Magnitude, a.k.a. norm, of the point."""
        return self._vct.mag

    @property
    def mag2(self):
        """Square of the magnitude, a.k.a. norm, of the point."""
        return self._vct.mag2

    def __eq__(self, another):
        """Equality of two points.

        Example
        -------
        >>> p1 = ...
        >>> p2 = ...
        >>> print  p1 == p2
        """
        return self._vct == another._vct

    def __ne__(self, another):
        """Non-equality of two points.

        Example
        -------
        >>> p1 = ...
        >>> p2 = ...
        >>> print  p1 != p2
        """
        return self._vct != another._vct

    ## operations
    def __iadd__(self, vector):
        """Add vector to point.

        Example
        -------
        >>> p  = Point3D  ( ... )
        >>> v  = Vector3D ( ... )
        >>> p += v
        """

        if isinstance(vector, Vector3D):
            self._vct += vector
            return self
        return NotImplemented

    def __isub__(self, vector):
        """Subtract vector from point.

        Example
        -------
        >>> p  = Point3D  ( ... )
        >>> v  = Vector3D ( ... )
        >>> p -= v
        """

        if isinstance(vector, Vector3D):
            self._vct -= vector
            return self
        return NotImplemented

    ## operations
    def __add__(self, vector):
        """Addition of point and vector.

        point + vector  --> point
        """
        ## point + vector = point
        if isinstance(vector, Vector3D):
            newpoint = self.copy()
            newpoint += vector
            return newpoint

        return NotImplemented

    def __sub__(self, other):
        """Subtraction for points and vectors.

        1. point - point   -->  vector
        2. point - vector  -->  point
        """
        ## point - point  --> vector
        if isinstance(other, Point3D):
            return self._vct - other._vct

        ## point - vector --> point
        if isinstance(other, Vector3D):
            newpoint = self.copy()
            newpoint -= other
            return newpoint

        return NotImplemented

    def __radd__(self, other):
        """Right addition of vector and point."""
        return self + other

    def __repr__(self):
        """Class representation."""
        return "<Point3D(x={0},y={1},z={2})>".format(self.x, self.y, self.z)

    def __str__(self):
        """Simple class representation."""
        return "Point3D(x={0},y={1},z={2})".format(self.x, self.y, self.z)

    def __eq__(self, other):
        """Equality  criteria for two points.

        Example
        -------
        >>> point1 = ...
        >>> point2 = ...
        >>> point1 == point2
        """
        if isinstance(other, Point3D):
            return self._vct == other._vct
        elif isinstance(other, (float, int)):
            return self._vct == other
        return NotImplemented

    def __ne__(self, other):
        """Non-equality  crietria for two points.

        Example
        -------
        >>> point1 = ...
        >>> point2 = ...
        >>> point1 != point2
        """
        return not (self == other)

    def on_line(self, line):
        """In the point on the line?"""
        v = line.point - self
        return 0 == v.mag2 or line.iscollinear(v)

    def on_plane(self, plane):
        """Is the point on the plane?"""
        v = plane.point - self
        return 0 == v.mag2 or plane.normal.isperpendicular(v)

    def distance(self, other):
        """Get 'distance' between point  and other object."""
        if isinstance(other, Point3D):
            return abs(self - other)
        elif isinstance(other, (Line3D, Plane3D)):
            return other.distance(self)
        raise NotImplementedError("Distance from Point3D to %s is not defined" % other)


# =============================================================================
# 3D-Line
# =============================================================================
class Line3D(object):
    """Line in 3D space.
    Line is defined by a point on line and by the direction.
    """

    def __init__(self, point=Point3D(0, 0, 0), vector=Vector3D(0, 0, 1)):
        if not isinstance(point, Point3D):
            raise NotImplementedError("Line3D: invalid ``point'' argument")
        if not isinstance(vector, Vector3D):
            raise NotImplementedError("Line3D: invalid ``vector'' argument")

        if 0 == vector.mag2:
            raise ValueError("Line3D: ``vector'' must not be zero!")

        self.point = point.copy()
        self.direction = vector.copy()

    @classmethod
    def from_points(cls, point1, point2):
        """Create the line from two points.

        Parameters
        ----------
        point1 : the first point on the line
        point2 : the second point on the line

        """
        return cls(point1, point2 - point1)

    @classmethod
    def from_line(cls, line):
        """Create the line from anoother line."""
        return cls(line.point, line.direction)

    def __repr__(self):
        """Simple class representation."""
        return "<Line3D({0},{1})>".format(self.point, self.direction)

    def __str__(self):
        """Simple class representation."""
        return "Line3D({0},{1})".format(self.point, self.direction)

    def copy(self):
        """Make a copy of this line."""
        return Line3D(self.point.copy(), self.direction.copy())

    ## is the point on line ?
    def __contains__(self, point):
        """Is the point on the line?

        Example
        -------
        >>> line  = ...
        >>> point = ...
        >>> if point in line : ...
        """
        return point.on_line(self)

    def iscollinear(self, other):
        """Is the line collinear to another line or vector?"""
        if isinstance(other, Line3D):
            other = other.direction
        elif isinstance(other, Vector3D):
            pass
        else:
            raise TypeError("Line3D.iscollinear: invalid ``other'' argument")
        ##
        return self.direction.iscollinear(other)

    def __eq__(self, other):
        """Equality  criteria for two lines.

        Example
        -------
        >>> line1 = ...
        >>> line2 = ...
        >>> print line1 == line2
        """
        return other.point in self and self.iscollinear(other)

    def __ne__(self, other):
        """Nonequality  criteria for two lines.

        Example
        -------
        >>> line1 = ...
        >>> line2 = ...
        >>> print line1 != line2
        """
        return not (self == other)

    def on_plane(self, plane):
        """Is the line on the plane?"""
        return self.point.on_plane(plane) and self.direction.isperpendicular(
            plane.normal
        )

    def point_on_line(self, mu):
        """Get the point on line, that  corresponds to parameter 'mu'.

        Example
        -------
        >>> line = ...
        >>> p0 = line.point_on_line ( 0 ) ## should be equal to line.point
        >>> p1 = line.point_on_line ( 1 ) ## shodul be equal to line.point+line.vector
        """
        return self.point + mu * self.direction

    def closest_point_param(self, point):
        """Get the 'mu' parameter of the point on line that is closest to the given point in space.

        Example
        -------
        >>> line  = ...
        >>> point = ...
        >>> mu    = line.closest_point_param ( point )
        """
        return (self.direction * (self.point - point)) / self.direction.mag2

    def closest_points_params(self, line):
        """Get the 'mu' parameters of two points on lines that is closest to each other.

        Example
        -------
        >>> line1 = ...
        >>> line2 = ...
        >>> mu1 , mu2  = line1.closest_points_param ( line2 )
        """
        a00 = self.direction.mag2
        a01 = self.direction * line.direction
        a10 = -a01
        a11 = -line.direction.mag2

        deti = a00 * a11 - a01 * a10

        ## two lines are actually parallel
        if isequal(deti, 0, scale=abs(a00 * a11)):
            return 0.0, line.closest_point_param(self.point)

        detinv = 1.0 / deti

        v = line.point - self.point

        b0 = v * self.direction
        b1 = v * line.direction

        mu1 = (a11 * b0 - a10 * b1) * detinv
        mu2 = (-a01 * b0 + a00 * b1) * detinv

        return mu1, mu2

    def closest_point(self, point):
        """Get the point on line that is closest to the given point in space.

        Example
        -------
        >>> line  = ...
        >>> point = ...
        >>> closest  = line.closest_point ( point )
        """
        return self.point_on_line(self.closest_point_param(point))

    def closest_points(self, line):
        """Get two points on lines that is closest to each other.

        Example
        -------
        >>> line1 = ...
        >>> line2 = ...
        >>> p1 , p2  = line1.closest_points ( line2 )
        """
        mu1, mu2 = self.closest_points_params(line)
        return self.point_on_line(mu1), line.point_on_line(mu2)

    def distance(self, other):
        """Get 'distance' between line and other object.

        Examples
        --------
        >>> line  = ...
        >>> point = ...
        >>> line2 = ...
        >>> plane = ...
        >>> print line.distance ( line2 )
        >>> print line.distance ( point )
        >>> print line.distance ( plane )
        """
        if isinstance(other, Point3D):
            if other in self:
                return 0
            return abs(other - self.closest_point(other))
        elif isinstance(other, Line3D):
            if other == self:
                return 0
            p1, p2 = self.closest_points(other)
            return abs(p1 - p2)
        elif isinstance(other, Plane3D):
            return other.distance(self)

        raise NotImplementedError("Distance from Line3D to %s is not defined" % other)

    def angle(self, other):
        """Get the 'angle' between a line and other object."""
        if isinstance(other, Line3D):
            return self.direction.angle(other.direction)
        elif isinstance(other, Plane3D):
            return self.direction.angle(other.normal)

        raise NotImplementedError(
            "Angle between a Line3D and a %s is not defined" % other
        )

    def intersect(self, other):
        """Get the 'intersection' between a line and other object."""

        if isinstance(other, Point3D):
            if other in self:
                return other
            else:
                return None

        elif isinstance(other, Line3D):
            dist = self.distance(other)

            if isequal(dist, 0.0):
                return self.closest_points(other)[0]
            else:
                return None

        elif isinstance(other, Plane3D):
            if self.direction.isperpendicular(other.normal):
                if self.on_plane(other):
                    return self
                else:
                    return None
            else:
                nom = other.normal * (other.point - self.point)
                den = other.normal * (self.direction)
                mu = nom / den

                return self.point_on_line(mu)

        raise NotImplementedError(
            "Intersection between a Line3D and a %s is not defined" % other
        )


# =============================================================================
# 3D-Plane
# =============================================================================
class Plane3D(object):
    """Plane in 3D space, defined by a point on plane and by the normal vector."""

    def __init__(self, point=Point3D(0, 0, 0), normal=Vector3D(0, 0, 1)):
        if not isinstance(point, Point3D):
            raise NotImplementedError("Plane3D: invalid ``point'' argument")
        if not isinstance(normal, Vector3D):
            raise NotImplementedError("Plane3D: invalid ``normal'' argument")

        if 0 == normal.mag2:
            raise ValueError("Plane3D: ``normal'' must not be zero!")

        self.point = point.copy()
        self.normal = normal.copy()

    @classmethod
    def from_points(cls, point1, point2, point3):
        """Create the plane from three  points.

        Parameters
        ----------
        point1 : the first  point on the plane
        point2 : the second point on the plane
        point3 : the third  point on the plane
        """
        v21 = point2 - point1
        v31 = point3 - point1
        return cls(point1, v21.cross(v31))

    @classmethod
    def from_line_and_point(cls, line, point):
        """Create the plane from the line and point.

        Parameters
        ----------
        line   : the line in plane
        point  : the point in plane
        """
        return cls(point, line.direction.cross(point - line.point))

    ## is the point or line on plane ?
    def __contains__(self, obj):
        """Is the point or line on the line?

        Parameters
        ----------
        object : the point or line to be tested

        Example
        -------
        >>> line  = ...
        >>> point = ...
        >>> plane = ...
        >>> if point in plane : ...
        >>> if line  in plane : ...
        """
        return obj.on_plane(self)

    def isparallel(self, other):
        """Is the plane collinear to another line, vector or plane?"""
        if isinstance(other, Line3D):
            return self.normal.isperpendicular(other.direction)
        elif isinstance(other, Vector3D):
            return self.normal.isperpendicular(other)
        elif isinstance(other, Plane3D):
            return self.normal.iscollinear(other.normal)
        else:
            raise TypeError("Plane3D.isparallel: invalid ``other'' argument")

    def __repr__(self):
        """Simple class representation."""
        return "<Plane3D({0},{1})>".format(self.point, self.normal)

    def __str__(self):
        """Simple class representation."""
        return "Plane3D({0},{1})".format(self.point, self.normal)

    def copy(self):
        """Make a copy of this plane."""
        return Plane3D(self.point.copy(), self.normal.copy())

    def __eq__(self, other):
        """Equality  criteria for two planes.

        Example
        -------
        >>> plane1 = ...
        >>> plane2 = ...
        >>> print plane1 == plane2
        """
        return other.point in self and self.normal.iscollinear(other.normal)

    def __ne__(self, other):
        """Nonequality  criteria for two planes.

        Example
        -------
        >>> plane1 = ...
        >>> plane2 = ...
        >>> print plane1 != plane2
        """
        return not (self == other)

    def distance(self, other):
        """Get 'distance' between plane and other object.

        Examples
        --------
        >>> line  = ...
        >>> point = ...
        >>> line2 = ...
        >>> plane = ...
        >>> print line.distance ( line2 )
        >>> print line.distance ( point )
        >>> print line.distance ( plane )
        """
        try:
            if isinstance(other, Point3D):
                if other in self:
                    return 0
                v = other - self.point
                return abs(v * self.normal) / self.normal.mag

            elif self.isparallel(other):
                return self.distance(other.point)

            else:
                return 0

        except TypeError:
            raise NotImplementedError(
                "Distance from Line3D to %s is not defined" % other
            )

    def angle(self, other):
        """Get the 'angle' between a plane and other object."""
        if isinstance(other, Line3D):
            return self.normal.angle(other.direction)
        elif isinstance(other, Plane3D):
            return self.normal.angle(other.normal)

        raise NotImplementedError(
            "Angle between a Line3D and a %s is not defined" % other
        )

    def intersect(self, other):
        """Get the 'intersection' between a line and other object"""

        if isinstance(other, Point3D):
            if other in self:
                return other
            else:
                return None

        elif isinstance(other, Line3D):
            return other.intersect(self)

        elif isinstance(other, Plane3D):
            if self.isparallel(other):
                return None
            else:
                N1, N2 = self.normal, other.normal
                d1, d2 = N1 * self.point._vct, N2 * other.point._vct

                det = N1 * N1 * N2 * N2 - (N1 * N2) ** 2
                deti = 1.0 / det

                c1 = (d1 * N2 * N2 - d2 * N1 * N1) * deti
                c2 = (d1 * N2 * N2 - d2 * N1 * N1) * deti

                point = Point3D.fromvector(c1 * N1 + c2 * N2)
                direction = N1.cross(N2)

                return Line3D(point, direction)

        raise NotImplementedError(
            "Intersection between a Plane3D and a %s is not defined" % other
        )


# ===========================================================================
def distance(obj1, obj2):
    """Calculate 3D-distance between two geometry objects (points, lines, planes).

    Examples
    --------
    >>> line1  = ...
    >>> line2  = ...
    >>> point1 = ...
    >>> point2 = ...
    >>> plane1 = ...
    >>> plane2 = ...
    >>> for o1 in ( line1, line2 , point1 , point2 , plane1 , plane2 ) :
    ...    for o2 in ( line1, line2 , point1 , point2 , plane1 , plane2 ) :
    ...        print distance ( o1 , o2 )
    """
    return obj1.distance(obj2)


# ===========================================================================
if "__main__" == __name__:
    v = Vector3D(1, 2, 3)
    p = Point3D(1, 2, 3)
    p += 0.5 * v
    p -= 0.5 * v
    p, p + v * 4, p - (p + v)

    line = Line3D(p, v)
    p in line

    line = Line3D.from_points(p, Point3D(3, 2, 1))
    p in line

    plane = Plane3D(p, v)
    line in plane

    plane = Plane3D.from_points(p, Point3D(3, 2, 1), Point3D(0, 0, 3))
    p in plane
